/*
Module : CFGSmsParser.H
Purpose: Parser FGSMS token 
Created: Francesco Germinara (c) 2001

Copyright (c) 2001 by SofTech di Germinara Francesco
All rights reserved.
*/

/////////////////////////////// Defines ///////////////////////////////////////
#ifndef __CFGSmsParser_H__
#define __CFGSmsParser_H__

/////////////////////////////// Classes ///////////////////////////////////////
//Token
class CFGSmsToken
{
public:
  CString m_strToken;
  int m_eType;
  CFGSmsToken(CString s, int eType,CString funct(CStringArray& parms,CStringArray& options)){m_strToken = s; m_eType = eType; m_pFunct=&funct; };
  CString (__cdecl *m_pFunct)(CStringArray& parms,CStringArray& options); 
};

class CFGSmsTokens
{
private:
  CPtrArray m_paTokens;
public:
  void Add(CString s, int eType, CString funct(CStringArray& parms,CStringArray& options)){ m_paTokens.Add(new CFGSmsToken(s,eType,&funct));};
  int GetSize() { return m_paTokens.GetSize();};
  CFGSmsToken *GetSmsToken(int i){CFGSmsToken *ptr=(CFGSmsToken *)m_paTokens.GetAt(i); return ptr;};
  CFGSmsToken *GetSmsToken(CString cmd){CFGSmsToken *ptr=NULL;
   for(int i=0; i < m_paTokens.GetSize(); i++){
    ptr=(CFGSmsToken *)m_paTokens.GetAt(i);
    if(ptr->m_strToken==cmd)
     return ptr;
   }
   return NULL;
  };

  void RemoveAll(){ 
   for(int i=0; i <m_paTokens.GetSize();i++){
	if(GetSmsToken(i) != NULL) delete GetSmsToken(i);
   };
   m_paTokens.RemoveAll();
  };
};

class CFGSmsParser
{
private:
 CString m_strStringToParser;   
 CFGSmsTokens m_strArrayOfToken;

 int m_nCommands;


 CString m_strLastError;

 char SMS[160]; //max SMS command
 char *ptrSMS; //pointer to SMS command 

public:
  enum eToken {eStartCmd,eDelimiter,eCM,eGM,ePM,eRM,eUI,eVI,eBC,eNM,eDM,eST,eSM,eSB,eRI,eWhiteSpace,eString,eNumeric,eTimeSeparator,eOption,eComment,eError,eEnd };
  int GetToken(); 
  void SetStrToParser(CString strToParser);
  CString GetLastError(){ return m_strLastError;};
  void CleanUp();

  //Constructors / Destructors
  CFGSmsParser(CString strToParser=_T("")); //Standard contructor
  ~CFGSmsParser(); //Standard contructor

  CString cmdToExecute;
  CStringArray cmdParams;
  CStringArray cmdOptions;
  CFGSmsParser *m_pScheduleCmd;

protected:
 int LookUp(int type,int startFrom,CString& token);
 int LookUp(CString strIDcmd);
 BOOL IsEnd(int &tokenType);
 BOOL IsWhiteSpace(int &tokenType);
 BOOL IsDelimiter(int &tokenType);
 BOOL IsComment(int &tokenType);
 BOOL IsOption(int &tokenType);
 BOOL IsCommand(int &tokenType);
 int IsNumeric(CString str);

 int GetCommand();
 int GetParameter();
 int GetOption();

 

public:
 CString Parser(BOOL debugInfo=FALSE);
};
#endif 

